-------------------------------------------------------------------------------
--                      Transactional Accounts (body)                        --
--                                                                           --
--                        by Simon Schule,                                   --
--             Last modified 26.09.2000, by Xavier Caron                     --
--                      e-mail:  xavier.caron@epfl.ch                        --
--                                                                           --
--                                                                           --
--  The Transactional Account class contains the type for an user account of --
--  the auction system.                                                      --
-------------------------------------------------------------------------------

with Operations.Creation; use Operations.Creation;
with Operations.Loading; use Operations.Loading;
with Operations.Saving; use Operations.Saving;
with Operations.Normal; use Operations.Normal;
with Operations.Deletion; use Operations.Deletion;

with Memory_Objects; use Memory_Objects;
with Memory_Objects.Concrete; use Memory_Objects.Concrete;

with Locks; use Locks;
with Procedural_Transaction_Interface;
use Procedural_Transaction_Interface;

with Streams; use Streams;

with Ada.Unchecked_Conversion;
with Ada.Unchecked_Deallocation;

with Accounts; use Accounts;

--  For debugging
with Debug; use Debug;

package body Transactional_Accounts is

   -------------------------------------------------------------------
   --  type conversion

   function To_Data_Ref is
      new Ada.Unchecked_Conversion (Account_Ref, Data_Ref);

   pragma Warnings (Off);
   function To_Account_Ref is
      new Ada.Unchecked_Conversion (Data_Ref, Account_Ref);
   pragma Warnings (on);

   -------------------------------------------------------------------
   --  operation for account_type creation

   type Account_Creation_Type is new Creation_Operation_Type with null record;

   function Create (Operation : Account_Creation_Type) return
     Data_Ref;

   function Create (Operation : Account_Creation_Type) return
     Data_Ref is
      Result : Account_Ref := new Account_Type;
   begin
      Put_Line ("Created Account Instance");
      return To_Data_Ref (Result);
   end Create;

   -------------------------------------------------------------------
   --  operation for account_type loading

   type Account_Loading_Type is new Loading_Operation_Type with null record;

   procedure Load (Operation   : Account_Loading_Type;
                   Data_Object : Data_Ref;
                   Stream      : Abstract_Stream_Ref);

   procedure Create_And_Load (Operation : Account_Loading_Type;
                              Data_Object : out Data_Ref;
                              Stream : Abstract_Stream_Ref);

   procedure Create_And_Load (Operation : Account_Loading_Type;
                              Data_Object : out Data_Ref;
                              Stream : Abstract_Stream_Ref) is
      Result : Account_Ref := new Account_Type;
   begin
      Account_Type'Read (Stream, Result.all);
      Data_Object := To_Data_Ref (Result);
      Put_Line ("Created and Loaded Account Instance");
   end Create_And_Load;

   procedure Load (Operation   : Account_Loading_Type;
                   Data_Object : Data_Ref;
                   Stream      : Abstract_Stream_Ref) is
      Tmp : Account_Ref := To_Account_Ref (Data_Object);
   begin
      Account_Type'Read (Stream, Tmp.all);
      Put_Line ("Loaded Account Instance");
   end Load;

   -------------------------------------------------------------------
   --  operation for account_type saving

   type Account_Saving_Type is new Saving_Operation_Type with null record;

   procedure Save (Operation   : Account_Saving_Type;
                   Data_Object : Data_Ref;
                   Stream      : Abstract_Stream_Ref);

   procedure Save (Operation   : Account_Saving_Type;
                   Data_Object : Data_Ref;
                   Stream      : Abstract_Stream_Ref) is
   begin
      Account_Type'Write (Stream, To_Account_Ref (Data_Object).all);
   end Save;

   -------------------------------------------------------------------
   --  operation for account_type deleting

   procedure Free_Account is new
     Ada.Unchecked_Deallocation (Account_Type'Class, Account_Ref);

   type Account_Deletion_Type is new Deletion_Operation_Type with null record;

   procedure Delete (Operation : Account_Deletion_Type;
                     Data_Object : in out Data_Ref);

   procedure Delete (Operation : Account_Deletion_Type;
                     Data_Object : in out Data_Ref) is
      Tmp : Account_Ref := To_Account_Ref (Data_Object);
   begin
      Free_Account (Tmp);
      Data_Object := null;
   end Delete;

   -------------------------------------------------------------------
   --  account locks

   type Kind_Type is (Deposit, Withdraw, Balance,
                      Transactional_Balance);

   type Account_Lock_Type (Kind : Kind_Type) is new Lock_Type with record
      Amount : Natural;
      Balance : Natural;
   end record;

   type Account_Lock_Ref is access all Account_Lock_Type'Class;

   function Is_Update (Lock : Account_Lock_Type) return Boolean;

   function Is_Compatible (Lock : Account_Lock_Type;
                           Other_Lock : Account_Lock_Type)
     return Boolean;

   function Is_Update (Lock : Account_Lock_Type) return Boolean is
   begin
      return Lock.Kind = Deposit or Lock.Kind = Withdraw;
   end Is_Update;

   function Is_Compatible (Lock : Account_Lock_Type;
                           Other_Lock : Account_Lock_Type)
                          return Boolean is
   begin
      case Lock.Kind is
         when Deposit =>
            return not (Other_Lock.Kind = Balance);
         when Withdraw =>
            return Other_Lock.Kind = Withdraw or
              Other_Lock.Kind = Transactional_Balance;
         when Balance =>
            return Other_Lock.Kind = Balance or
              Other_Lock.Kind = Transactional_Balance;
         when Transactional_Balance =>
            return True;
      end case;
   end Is_Compatible;

   Deposit_Lock : Account_Lock_Ref :=
     new Account_Lock_Type (Deposit);
   Withdraw_Lock : Account_Lock_Ref :=
     new Account_Lock_Type (Withdraw);
   Balance_Lock : Account_Lock_Ref :=
     new Account_Lock_Type (Balance);
   Transactional_Balance_Lock : Account_Lock_Ref :=
     new Account_Lock_Type (Transactional_Balance);

   -------------------------------------------------------------------
   --  operations for Deposit

   type Deposit_Operation_Type is new Normal_Operation_Type with record
      Value : Natural;
   end record;

   procedure Do_Operation (Operation   : in out Deposit_Operation_Type;
                           Data_Object : in Data_Ref);

   procedure Undo_Operation (Operation   : in out Deposit_Operation_Type;
                             Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : in Deposit_Operation_Type)
      return Lock_Ref;

   function Is_Update (Operation : in Deposit_Operation_Type)
      return Boolean;

   procedure Do_Operation (Operation   : in out Deposit_Operation_Type;
                           Data_Object : in Data_Ref) is
   begin
      Deposit (To_Account_Ref (Data_Object).all, Operation.Value);
   end Do_Operation;

   procedure Undo_Operation (Operation   : in out Deposit_Operation_Type;
                               Data_Object : in Data_Ref) is
   begin
      Withdraw (To_Account_Ref (Data_Object).all, Operation.Value);
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : in Deposit_Operation_Type)
                             return Lock_Ref is
   begin
      return Lock_Ref (Deposit_Lock);
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : in Deposit_Operation_Type)
                      return Boolean is
   begin
      return True;
   end Is_Update;

   -------------------------------------------------------------------
   --  operations for Withdraw

   type Withdraw_Operation_Type is new Normal_Operation_Type with record
      Value : Natural;
   end record;

   procedure Do_Operation (Operation   : in out Withdraw_Operation_Type;
                           Data_Object : in Data_Ref);

   procedure Undo_Operation (Operation   : in out Withdraw_Operation_Type;
                             Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : in Withdraw_Operation_Type)
      return Lock_Ref;

   function Is_Update (Operation : in Withdraw_Operation_Type)
      return Boolean;

   procedure Do_Operation (Operation   : in out Withdraw_Operation_Type;
                           Data_Object : in Data_Ref) is
   begin
      Withdraw (To_Account_Ref (Data_Object).all, Operation.Value);
   end Do_Operation;

   procedure Undo_Operation (Operation   : in out Withdraw_Operation_Type;
                               Data_Object : in Data_Ref) is
   begin
      Deposit (To_Account_Ref (Data_Object).all, Operation.Value);
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : in Withdraw_Operation_Type)
                             return Lock_Ref is
   begin
      return Lock_Ref (Withdraw_Lock);
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : in Withdraw_Operation_Type)
                      return Boolean is
   begin
      return True;
   end Is_Update;

   -------------------------------------------------------------------
   --  operations for Balance

   type Balance_Operation_Type is
     new Normal_Operation_Type with record
      Balance : Natural;
     end record;

   procedure Do_Operation (Operation : in out Balance_Operation_Type;
                           Data_Object : in Data_Ref);

   procedure Undo_Operation
     (Operation : in out Balance_Operation_Type;
      Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : in Balance_Operation_Type)
      return Lock_Ref;

   function Is_Update (Operation : in Balance_Operation_Type)
      return Boolean;

   procedure Do_Operation (Operation : in out Balance_Operation_Type;
                           Data_Object : in Data_Ref) is
   begin
      Operation.Balance := Balance (To_Account_Ref (Data_Object).all);
   end Do_Operation;

   procedure Undo_Operation
     (Operation : in out Balance_Operation_Type;
      Data_Object : in Data_Ref) is
   begin
      null;
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : in Balance_Operation_Type)
                             return Lock_Ref is
   begin
      return Lock_Ref (Balance_Lock);
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : in Balance_Operation_Type)
                      return Boolean is
   begin
      return False;
   end Is_Update;

   -------------------------------------------------------------------
   --  operations for Transactional_Balance

   type Transactional_Balance_Operation_Type is
     new Normal_Operation_Type with record
        Balance : Natural;
     end record;

   procedure Do_Operation
     (Operation : in out Transactional_Balance_Operation_Type;
      Data_Object : in Data_Ref);

   procedure Undo_Operation
     (Operation   : in out Transactional_Balance_Operation_Type;
      Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock
     (Operation : in Transactional_Balance_Operation_Type)
     return Lock_Ref;

   function Is_Update
     (Operation : in Transactional_Balance_Operation_Type)
     return Boolean;

   procedure Do_Operation
     (Operation   : in out Transactional_Balance_Operation_Type;
      Data_Object : in Data_Ref) is
   begin
      Operation.Balance := Balance (To_Account_Ref (Data_Object).all);
   end Do_Operation;

   procedure Undo_Operation
     (Operation   : in out Transactional_Balance_Operation_Type;
      Data_Object : in Data_Ref) is
   begin
      null;
   end Undo_Operation;

   function Get_Kind_Of_Lock
     (Operation : in Transactional_Balance_Operation_Type)
     return Lock_Ref is
   begin
      return Lock_Ref (Transactional_Balance_Lock);
   end Get_Kind_Of_Lock;

   function Is_Update
     (Operation : in Transactional_Balance_Operation_Type)
     return Boolean is
   begin
      return False;
   end Is_Update;

   -------------------------------------------------------------------
   --  operations of the proxy

   procedure Deposit (A : in out Transactional_Account_Type;
                      Value : in Natural) is
      My_Deposit_Operation : Deposit_Operation_Type;
   begin
      My_Deposit_Operation.Value := Value;
      Atomic_Do (A.My_Atomic_Call.all, My_Deposit_Operation);
   end Deposit;

   procedure Withdraw (A : in out Transactional_Account_Type;
                       Value : in Natural) is
      My_Withdraw_Operation : Withdraw_Operation_Type;
   begin
      My_Withdraw_Operation.Value := Value;
      Atomic_Do (A.My_Atomic_Call.all, My_Withdraw_Operation);
   end Withdraw;

   function Balance (A : Transactional_Account_Type) return Natural is
      My_Balance_Operation : Transactional_Balance_Operation_Type;
   begin
      Atomic_Do (A.My_Atomic_Call.all, My_Balance_Operation);
      return My_Balance_Operation.Balance;
   end Balance;

   function Create_Account (Params : in Non_Volatile_Params_Type'Class) return
     Transactional_Account_Ref is
      Result : Transactional_Account_Ref := new Transactional_Account_Type;
      My_Creation_Operation : Account_Creation_Type;
      My_Loading_Operation : Account_Loading_Type;
      My_Saving_Operation : Account_Saving_Type;
      My_Deletion_Operation : Account_Deletion_Type;
   begin
      Result.My_Memory_Object := Concrete_Memory_Object_Ref
        (Create (Params,
                 My_Creation_Operation, My_Loading_Operation,
                 My_Saving_Operation, My_Deletion_Operation));
      Result.My_Atomic_Call := new Atomic_Call_Type (Result.My_Memory_Object);
      return Result;
   end Create_Account;

   function Restore_Account (Params : in Non_Volatile_Params_Type'Class) return
     Transactional_Account_Ref is
      Result : Transactional_Account_Ref := new Transactional_Account_Type;
      My_Loading_Operation : Account_Loading_Type;
      My_Saving_Operation : Account_Saving_Type;
      My_Deletion_Operation : Account_Deletion_Type;
   begin
      Result.My_Memory_Object := Concrete_Memory_Object_Ref
        (Restore (Params,
                  My_Loading_Operation,
                  My_Saving_Operation, My_Deletion_Operation));
      Result.My_Atomic_Call := new Atomic_Call_Type (Result.My_Memory_Object);
      return Result;
   end Restore_Account;

end Transactional_Accounts;
