-------------------------------------------------------------------------------
--                        Transactional Auctions (body part)                 --
--                                                                           --
--                  by Xavier Caron, last modified 18.10.2000                --
--                                                                           --
--                      e-mail:  xavier.caron@epfl.ch                        --
--                                                                           --
--                                                                           --
--  The Transactional Auction class contains the type auction of the         --
--  auction system.                                                          --
-------------------------------------------------------------------------------
with Operations.Creation; use Operations.Creation;
with Operations.Loading; use Operations.Loading;
with Operations.Saving; use Operations.Saving;
with Operations.Deletion; use Operations.Deletion;
with Operations.Normal; use Operations.Normal;

with Memory_Objects; use Memory_Objects;
with Memory_Objects.Concrete; use Memory_Objects.Concrete;

with Locks; use Locks;
with Locks.Read_Write; use Locks.Read_Write;

with Procedural_Transaction_Interface;
use Procedural_Transaction_Interface;

with Streams; use Streams;

with Ada.Finalization;
with Ada.Unchecked_Conversion;
with Ada.Unchecked_Deallocation;

with Auctions; use Auctions;

--  For debugging
with Debug; use Debug;
with Debug_Auction;
package body Transactional_Auctions is

   My_Read_Lock : Lock_Ref := Lock_Ref (Read_Lock);
   --  My_Write_Lock : Lock_Ref := Lock_Ref (Write_Lock);

   -------------------------------------------------------------------
   --  type conversion

   function To_Data_Ref is
      new Ada.Unchecked_Conversion (Auction_Ref, Data_Ref);

   pragma Warnings (OFF);
   function To_Auction_Ref is
      new Ada.Unchecked_Conversion (Data_Ref, Auction_Ref);
   pragma Warnings (ON);

   -------------------------------------------------------------------
   --  operation for auction_type creation

   type Auction_Creation_Type (N, D : Natural) is new Creation_Operation_Type
     with record
        Name : String (1 .. N);
        Description : String (1 .. D);
        Initial_Bid : Natural;
        Seller_Task_Id : Task_Id;
     end record;

   function Create (Operation : Auction_Creation_Type)
                   return Data_Ref;

   function Create (Operation : Auction_Creation_Type)
                   return Data_Ref is
      Result : Auction_Ref := Create_Auction
        (Operation.Name, Operation.Description,
         Operation.Initial_Bid, Operation.Seller_Task_Id);
   begin
      Put_Line ("Created Auction Instance");
      return To_Data_Ref (Result);
   end Create;

   -------------------------------------------------------------------
   --  operation for auction_type loading

   type Auction_Loading_Type is new Loading_Operation_Type with null record;

   procedure Load (Operation : Auction_Loading_Type;
                   Data_Object : Data_Ref;
                   Stream : Abstract_Stream_Ref);

   procedure Create_And_Load (Operation : Auction_Loading_Type;
                              Data_Object : out Data_Ref;
                              Stream : Abstract_Stream_Ref);

   procedure Create_And_Load (Operation : Auction_Loading_Type;
                              Data_Object : out Data_Ref;
                              Stream : Abstract_Stream_Ref) is
      Result : Auction_Ref := new Auction_Type' (Auction_Type'Input (Stream));
   begin
      Data_Object := To_Data_Ref (Result);
      Put_Line ("Loaded Auction Instance");
   end Create_And_Load;

   procedure Free_Auction is new Ada.Unchecked_Deallocation
     (Auction_Type, Auction_Ref);

   procedure Load (Operation : Auction_Loading_Type;
                   Data_Object : Data_Ref;
                   Stream : Abstract_Stream_Ref) is
      Temp : Auction_Ref := To_Auction_Ref (Data_Object);
   begin
      Load_State (Temp.all, Stream);
      Put_Line ("Loaded Auction Instance");
   end Load;

   -------------------------------------------------------------------
   --  operation for auction_type saving

   type Auction_Saving_Type is new Saving_Operation_Type with null record;

   procedure Save (Operation : Auction_Saving_Type;
                   Data_Object : Data_Ref;
                   Stream : Abstract_Stream_Ref);

   procedure Save (Operation : Auction_Saving_Type;
                   Data_Object : Data_Ref;
                   Stream : Abstract_Stream_Ref) is
   begin
      Auction_Type'Output
        (Stream, To_Auction_Ref (Data_Object).all);
   end Save;

   -------------------------------------------------------------------
   --  operation for auction_type deleting

   type Auction_Deletion_Type is new Deletion_Operation_Type with null record;

   procedure Delete (Operation : Auction_Deletion_Type;
                     Data_Object : in out Data_Ref);

   procedure Delete (Operation : Auction_Deletion_Type;
                     Data_Object : in out Data_Ref) is
      Tmp : Auction_Ref := To_Auction_Ref (Data_Object);
   begin
      Free_Auction (Tmp);
      Data_Object := null;
   end Delete;

   -------------------------------------------------------------------
   --  operation for auction_type Bid

   type Bid_Operation_Type is new Normal_Operation_Type with record
      Old_Bid : Natural;
      New_Bid : Natural;
   end record;

   procedure Do_Operation (Operation : in out Bid_Operation_Type;
                           Data_Object : in Data_Ref);

   procedure Undo_Operation (Operation : in out Bid_Operation_Type;
                             Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : Bid_Operation_Type)
     return Lock_Ref;

   function Is_Update (Operation : Bid_Operation_Type)
     return Boolean;

   --  Body part

   procedure Do_Operation (Operation : in out Bid_Operation_Type;
                           Data_Object : in Data_Ref) is
   begin
      Operation.Old_Bid := Get_Current_Bid (To_Auction_Ref (Data_Object).all);
      Bid (To_Auction_Ref (Data_Object).all, Operation.New_Bid);
   end Do_Operation;

   procedure Undo_Operation
     (Operation   : in out Bid_Operation_Type;
      Data_Object : in Data_Ref) is
   begin
      --  Bid (To_Auction_Ref (Data_Object).all, Operation.Old_Bid);
      null;
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : Bid_Operation_Type)
     return Lock_Ref is
   begin
      return My_Read_Lock;
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : Bid_Operation_Type)
     return Boolean is
   begin
      return True;
   end Is_Update;

   -------------------------------------------------------------------
   --  operation for auction_type Is_Accepted

   type Is_Accepted_Operation_Type is new Normal_Operation_Type with record
      Result : Boolean;
   end record;

   procedure Do_Operation (Operation : in out Is_Accepted_Operation_Type;
                           Data_Object : in Data_Ref);

   procedure Undo_Operation (Operation : in out Is_Accepted_Operation_Type;
                             Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : Is_Accepted_Operation_Type)
     return Lock_Ref;

   function Is_Update (Operation : Is_Accepted_Operation_Type)
     return Boolean;

   --  Body part

   procedure Do_Operation (Operation : in out Is_Accepted_Operation_Type;
                           Data_Object : in Data_Ref) is
   begin
      Operation.Result := Is_Accepted (To_Auction_Ref (Data_Object).all);
   end Do_Operation;

   procedure Undo_Operation
     (Operation   : in out Is_Accepted_Operation_Type;
      Data_Object : in Data_Ref) is
   begin
      null;
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : Is_Accepted_Operation_Type)
     return Lock_Ref is
   begin
      return My_Read_Lock;
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : Is_Accepted_Operation_Type)
                      return Boolean is
   begin
      return False;
   end Is_Update;

   -------------------------------------------------------------------
   --  operation for auction_type Status

   type Status_Operation_Type is new Normal_Operation_Type with record
      Result : Auction_Status;
   end record;

   procedure Do_Operation (Operation : in out Status_Operation_Type;
                           Data_Object : in Data_Ref);

   procedure Undo_Operation (Operation : in out Status_Operation_Type;
                             Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : Status_Operation_Type)
     return Lock_Ref;

   function Is_Update (Operation : Status_Operation_Type)
     return Boolean;

   --  Body part

   procedure Do_Operation (Operation : in out Status_Operation_Type;
                           Data_Object : in Data_Ref) is
   begin
      Operation.Result := Status (To_Auction_Ref (Data_Object).all);
   end Do_Operation;

   procedure Undo_Operation
     (Operation   : in out Status_Operation_Type;
      Data_Object : in Data_Ref) is
   begin
      null;
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : Status_Operation_Type)
     return Lock_Ref is
   begin
      return My_Read_Lock;
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : Status_Operation_Type)
                      return Boolean is
   begin
      return False;
   end Is_Update;

   -------------------------------------------------------------------
   --  operation for auction_type Set_Status

   type Set_Status_Operation_Type is new Normal_Operation_Type with record
      New_Status : Auction_Status;
   end record;
   procedure Do_Operation (Operation : in out Set_Status_Operation_Type;
                           Data_Object : in Data_Ref);

   procedure Undo_Operation (Operation : in out Set_Status_Operation_Type;
                             Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : Set_Status_Operation_Type)
     return Lock_Ref;

   function Is_Update (Operation : Set_Status_Operation_Type)
     return Boolean;

   --  Body part

   procedure Do_Operation (Operation : in out Set_Status_Operation_Type;
                           Data_Object : in Data_Ref) is
   begin
      Set_Status (To_Auction_Ref (Data_Object).all, Operation.New_Status);
   end Do_Operation;

   procedure Undo_Operation
     (Operation   : in out Set_Status_Operation_Type;
      Data_Object : in Data_Ref) is
   begin
      --  we don't want to undo this!!!!
      null;
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : Set_Status_Operation_Type)
     return Lock_Ref is
   begin
      return My_Read_Lock;
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : Set_Status_Operation_Type)
     return Boolean is
   begin
      return True;
   end Is_Update;

   -------------------------------------------------------------------
   --  operation for auction_type get_name

   type String_Ref is access String;

   type String_Type is new Ada.Finalization.Controlled with record
      S : String_Ref;
   end record;

   procedure Finalize (S : in out String_Type);

   procedure Free is new
     Ada.Unchecked_Deallocation (String, String_Ref);

   procedure Finalize (S : in out String_Type) is
   begin
      Free (S.S);
   end Finalize;

   type Get_Name_Operation_Type is new Normal_Operation_Type with record
      Name : String_Type;
   end record;

   procedure Do_Operation (Operation : in out Get_Name_Operation_Type;
                           Data_Object : in Data_Ref);

   procedure Undo_Operation (Operation : in out Get_Name_Operation_Type;
                             Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : Get_Name_Operation_Type)
     return Lock_Ref;

   function Is_Update (Operation : Get_Name_Operation_Type)
     return Boolean;

   --  Body part

   procedure Do_Operation (Operation : in out Get_Name_Operation_Type;
                           Data_Object : in Data_Ref) is
   begin
      Operation.Name.S :=
        new String' (Get_Name (To_Auction_Ref (Data_Object).all));
   end Do_Operation;

   procedure Undo_Operation (Operation   : in out Get_Name_Operation_Type;
                             Data_Object : in Data_Ref) is
   begin
      null;
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : Get_Name_Operation_Type)
     return Lock_Ref is
   begin
      return My_Read_Lock;
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : Get_Name_Operation_Type)
     return Boolean is
   begin
      return False;
   end Is_Update;

   -------------------------------------------------------------------
   --  operation for auction_type get_description

   type Get_Description_Operation_Type is new Normal_Operation_Type with record
      Description : String_Type;
   end record;

   procedure Do_Operation (Operation : in out Get_Description_Operation_Type;
                           Data_Object : in Data_Ref);

   procedure Undo_Operation (Operation : in out Get_Description_Operation_Type;
                             Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : Get_Description_Operation_Type)
     return Lock_Ref;

   function Is_Update (Operation : Get_Description_Operation_Type)
     return Boolean;

   --  Body part

   procedure Do_Operation (Operation : in out Get_Description_Operation_Type;
                           Data_Object : in Data_Ref) is
   begin
      Operation.Description.S :=
        new String' (Get_Description (To_Auction_Ref (Data_Object).all));
   end Do_Operation;

   procedure Undo_Operation
     (Operation   : in out Get_Description_Operation_Type;
      Data_Object : in Data_Ref) is
   begin
      null;
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : Get_Description_Operation_Type)
     return Lock_Ref is
   begin
      return My_Read_Lock;
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : Get_Description_Operation_Type)
     return Boolean is
   begin
      return False;
   end Is_Update;

   -------------------------------------------------------------------
   --  operation for auction_type get_current_bid

   type Get_Current_Bid_Operation_Type is new Normal_Operation_Type with record
      Current_Bid : Natural;
   end record;

   procedure Do_Operation (Operation : in out Get_Current_Bid_Operation_Type;
                           Data_Object : in Data_Ref);

   procedure Undo_Operation (Operation : in out Get_Current_Bid_Operation_Type;
                             Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : Get_Current_Bid_Operation_Type)
     return Lock_Ref;

   function Is_Update (Operation : Get_Current_Bid_Operation_Type)
     return Boolean;

   --  Body part

   procedure Do_Operation (Operation : in out Get_Current_Bid_Operation_Type;
                           Data_Object : in Data_Ref) is
   begin
      Operation.Current_Bid := Get_Current_Bid
        (To_Auction_Ref (Data_Object).all);
   end Do_Operation;

   procedure Undo_Operation
     (Operation   : in out Get_Current_Bid_Operation_Type;
      Data_Object : in Data_Ref) is
   begin
      null;
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : Get_Current_Bid_Operation_Type)
     return Lock_Ref is
   begin
      return My_Read_Lock;
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : Get_Current_Bid_Operation_Type)
     return Boolean is
   begin
      return False;
   end Is_Update;

   -------------------------------------------------------------------
   --  operation for auction_type get_initial_bid

   type Get_Initial_Bid_Operation_Type is new Normal_Operation_Type with record
      Initial_Bid : Natural;
   end record;

   procedure Do_Operation (Operation : in out Get_Initial_Bid_Operation_Type;
                           Data_Object : in Data_Ref);

   procedure Undo_Operation (Operation : in out Get_Initial_Bid_Operation_Type;
                             Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : Get_Initial_Bid_Operation_Type)
     return Lock_Ref;

   function Is_Update (Operation : Get_Initial_Bid_Operation_Type)
     return Boolean;

   --  Body part

   procedure Do_Operation (Operation : in out Get_Initial_Bid_Operation_Type;
                           Data_Object : in Data_Ref) is
   begin
      Operation.Initial_Bid := Get_Initial_Bid
        (To_Auction_Ref (Data_Object).all);
   end Do_Operation;

   procedure Undo_Operation
     (Operation   : in out Get_Initial_Bid_Operation_Type;
      Data_Object : in Data_Ref) is
   begin
      null;
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : Get_Initial_Bid_Operation_Type)
     return Lock_Ref is
   begin
      return My_Read_Lock;
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : Get_Initial_Bid_Operation_Type)
     return Boolean is
   begin
      return False;
   end Is_Update;

   -------------------------------------------------------------------
   --  operation for auction_type get_seller_task_id

   type Get_Seller_Task_Id_Operation_Type is new
     Normal_Operation_Type with record
      Seller_Task_Id : Task_Id;
     end record;

   procedure Do_Operation
     (Operation : in out Get_Seller_Task_Id_Operation_Type;
      Data_Object : in Data_Ref);

   procedure Undo_Operation
     (Operation : in out Get_Seller_Task_Id_Operation_Type;
      Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : Get_Seller_Task_Id_Operation_Type)
     return Lock_Ref;

   function Is_Update (Operation : Get_Seller_Task_Id_Operation_Type)
     return Boolean;

   --  Body part

   procedure Do_Operation
     (Operation : in out Get_Seller_Task_Id_Operation_Type;
      Data_Object : in Data_Ref) is
   begin
      Operation.Seller_Task_Id := Get_Seller_Task_Id
        (To_Auction_Ref (Data_Object).all);
   end Do_Operation;

   procedure Undo_Operation
     (Operation   : in out Get_Seller_Task_Id_Operation_Type;
      Data_Object : in Data_Ref) is
   begin
      null;
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : Get_Seller_Task_Id_Operation_Type)
     return Lock_Ref is
   begin
      return My_Read_Lock;
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : Get_Seller_Task_Id_Operation_Type)
     return Boolean is
   begin
      return False;
   end Is_Update;

   -------------------------------------------------------------------
   --  operations of the proxy

   procedure Place_Bid (Auction : Transactional_Auction_Type;
                  New_Bid : Natural) is
      My_Bid_Operation : Bid_Operation_Type;
   begin
      My_Bid_Operation.New_Bid := New_Bid;
      Atomic_Do (Auction.My_Atomic_Call.all, My_Bid_Operation);
   end Place_Bid;

   function Is_Accepted (Auction : Transactional_Auction_Type)
                        return Boolean is
      Is_Accepted_Operation : Is_Accepted_Operation_Type;
   begin
      Atomic_Do (Auction.My_Atomic_Call.all, Is_Accepted_Operation);
      return Is_Accepted_Operation.Result;
   end Is_Accepted;

   function Status (Auction : Transactional_Auction_Type)
                    return Auction_Status is
      Status_Operation : Status_Operation_Type;
   begin
      Atomic_Do (Auction.My_Atomic_Call.all, Status_Operation);
      return Status_Operation.Result;
   end Status;

   procedure Set_Status (Auction : Transactional_Auction_Type;
                         Status : Auction_Status)
   is
      Set_Status_Operation : Set_Status_Operation_Type;
   begin
      Set_Status_Operation.New_Status := Status;
      Atomic_Do (Auction.My_Atomic_Call.all, Set_Status_Operation);
   end Set_Status;

   function Get_Name (Auction : Transactional_Auction_Type) return String is
      My_Get_Name_Operation : Get_Name_Operation_Type;
   begin
      Atomic_Do (Auction.My_Atomic_Call.all, My_Get_Name_Operation);
      return My_Get_Name_Operation.Name.S.all;
   end Get_Name;

   function Get_Description (Auction : Transactional_Auction_Type)
                            return String is
      My_Get_Description_Operation : Get_Description_Operation_Type;
   begin
      Atomic_Do (Auction.My_Atomic_Call.all, My_Get_Description_Operation);
      return My_Get_Description_Operation.Description.S.all;
   end Get_Description;

   function Get_Current_Bid (Auction : Transactional_Auction_Type)
                        return Natural is

      My_Get_Current_Bid_Operation :  Get_Current_Bid_Operation_Type;
   begin
      Atomic_Do (Auction.My_Atomic_Call.all, My_Get_Current_Bid_Operation);
      return My_Get_Current_Bid_Operation.Current_Bid;
   end Get_Current_Bid;

   function Get_Initial_Bid (Auction : Transactional_Auction_Type)
                            return Natural is
      My_Get_Initial_Bid_Operation :  Get_Initial_Bid_Operation_Type;
   begin
      Atomic_Do (Auction.My_Atomic_Call.all, My_Get_Initial_Bid_Operation);
      return My_Get_Initial_Bid_Operation.Initial_Bid;
   end Get_Initial_Bid;

   function Get_Seller_Task_Id (Auction : Transactional_Auction_Type)
                               return Task_Id is
      My_Get_Seller_Task_Id_Operation :  Get_Seller_Task_Id_Operation_Type;
   begin
      Atomic_Do (Auction.My_Atomic_Call.all, My_Get_Seller_Task_Id_Operation);
      return My_Get_Seller_Task_Id_Operation.Seller_Task_Id;
   end Get_Seller_Task_Id;

   function Create_Auction (Params : in Non_Volatile_Params_Type'Class;
                            Name : String;
                            Description : String;
                            Initial_Bid : Natural;
                            Seller_Task_Id : Task_Id)
                           return Transactional_Auction_Ref is
      Result : Transactional_Auction_Ref := new Transactional_Auction_Type;
      My_Loading_Operation : Auction_Loading_Type;
      My_Creation_Operation : Auction_Creation_Type (Name'Length,
                                                     Description'Length);
      My_Saving_Operation : Auction_Saving_Type;
      My_Deletion_Operation : Auction_Deletion_Type;
   begin
      Debug_Auction.Put_Line ("Entered create auction" & Name & " " & description);
      My_Creation_Operation.Name := Name;
      My_Creation_Operation.Description := Description;
      My_Creation_Operation.Initial_Bid := Initial_Bid;
      My_Creation_Operation.Seller_Task_Id := Seller_Task_Id;
      Result.My_Memory_Object := Concrete_Memory_Object_Ref
        (Create (Params,
                 My_Creation_Operation, My_Loading_Operation,
                 My_Saving_Operation, My_Deletion_Operation));
      Result.My_Atomic_Call := new Atomic_Call_Type (Result.My_Memory_Object);
      Debug_Auction.Put_Line ("finished create auction");
      return Result;
   end Create_Auction;

   function Restore_Auction (Params : in Non_Volatile_Params_Type'Class)
                            return Transactional_Auction_Ref is
      Result : Transactional_Auction_Ref := new Transactional_Auction_Type;
      My_Loading_Operation : Auction_Loading_Type;
      My_Saving_Operation : Auction_Saving_Type;
      My_Deletion_Operation : Auction_Deletion_Type;
   begin
      Result.My_Memory_Object := Concrete_Memory_Object_Ref
        (Restore (Params,
                  My_Loading_Operation,
                  My_Saving_Operation, My_Deletion_Operation));
      Result.My_Atomic_Call := new Atomic_Call_Type (Result.My_Memory_Object);
      return Result;
   end Restore_Auction;

end Transactional_Auctions;

