-------------------------------------------------------------------------------
--                   Transactional Users Directories (body)                  --
--                                                                           --
--                  by Xavier Caron, last modified 06.10.2000                --
--                                                                           --
--                      e-mail:  xavier.caron@epfl.ch                        --
--                                                                           --
--                                                                           --
--  The Transactional Users Directories class contains the type for          --
--  the user of the auction system.                                          --
-------------------------------------------------------------------------------
with Operations.Creation; use Operations.Creation;
with Operations.Loading; use Operations.Loading;
with Operations.Saving; use Operations.Saving;
with Operations.Deletion; use Operations.Deletion;
with Operations.Normal; use Operations.Normal;

with Memory_Objects; use Memory_Objects;
with Memory_Objects.Concrete; use Memory_Objects.Concrete;

with Locks; use Locks;

with Procedural_Transaction_Interface;
use Procedural_Transaction_Interface;
--  with Storage_Params; use Storage_Params;

with Streams; use Streams;

with Ada.Unchecked_Conversion;
with Ada.Unchecked_Deallocation;

with User_Directories; use User_Directories;

--  For debugging
with Debug; use Debug;


package body Transactional_User_Directories is

   -------------------------------------------------------------------
   --  type conversion

   function To_Data_Ref is
      new Ada.Unchecked_Conversion (User_Directory_Ref, Data_Ref);

   pragma Warnings (OFF);
   function To_User_Directory_Ref is
      new Ada.Unchecked_Conversion (Data_Ref, User_Directory_Ref);
   pragma Warnings (ON);

   -------------------------------------------------------------------
   --  operation for user_Directory_type creation

   type User_Directory_Creation_Type is new Creation_Operation_Type
     with null record;

   function Create (Operation : User_Directory_Creation_Type)
     return Data_Ref;

   function Create (Operation : User_Directory_Creation_Type)
     return Data_Ref is
      Result : User_Directory_Ref := new User_Directory_Type;
   begin
      Put_Line ("Created User Directory Instance");
      return To_Data_Ref (Result);
   end Create;

   -------------------------------------------------------------------
   --  operation for user_Directory_type loading

   type User_Directory_Loading_Type is
     new Loading_Operation_Type with null record;

   procedure Load (Operation : User_Directory_Loading_Type;
                   Data_Object : Data_Ref;
                   Stream : Abstract_Stream_Ref);

   procedure Create_And_Load (Operation : User_Directory_Loading_Type;
                              Data_Object : out Data_Ref;
                              Stream : Abstract_Stream_Ref);

   procedure Create_And_Load (Operation : User_Directory_Loading_Type;
                              Data_Object : out Data_Ref;
                              Stream : Abstract_Stream_Ref) is
      Result : User_Directory_Ref := new User_Directory_Type;
   begin
      Load_Directory (Result.all, Stream);
      Data_Object := To_Data_Ref (Result);
      Put_Line ("Loaded User Directory Instance");
   end Create_And_Load;

   procedure Free is new Ada.Unchecked_Deallocation
     (User_Directory_Type, User_Directory_Ref);

   procedure Load (Operation : User_Directory_Loading_Type;
                   Data_Object : Data_Ref;
                   Stream : Abstract_Stream_Ref) is
      Temp : User_Directory_Ref := To_User_Directory_Ref (Data_Object);
   begin
      Load_Directory (Temp.all, Stream);
      Put_Line ("Loaded User Directory Instance");
   end Load;

   -------------------------------------------------------------------
   --  operation for user_Directory_type saving

   type User_Directory_Saving_Type is
     new Saving_Operation_Type with null record;

   procedure Save (Operation : User_Directory_Saving_Type;
                   Data_Object : Data_Ref;
                   Stream : Abstract_Stream_Ref);

   procedure Save (Operation : User_Directory_Saving_Type;
                   Data_Object : Data_Ref;
                   Stream : Abstract_Stream_Ref) is
   begin
      Save_Directory (To_User_Directory_Ref (Data_Object).all, Stream);
   end Save;

   -------------------------------------------------------------------
   --  operation for user_Directory_type deleting

   type User_Directory_Deletion_Type is
     new Deletion_Operation_Type with null record;

   procedure Free_User_Directory is new
     Ada.Unchecked_Deallocation (User_Directory_Type, User_Directory_Ref);

   procedure Delete (Operation : User_Directory_Deletion_Type;
                     Data_Object : in out Data_Ref);

   procedure Delete (Operation : User_Directory_Deletion_Type;
                     Data_Object : in out Data_Ref) is
      Tmp : User_Directory_Ref := To_User_Directory_Ref (Data_Object);
   begin
      Free_User_Directory (Tmp);
      Data_Object := null;
   end Delete;

   type Operation_Type is
     (Register, Get_All_Users, Get_User);

   type User_Directory_Lock_Type (Kind : Operation_Type)
     is new Lock_Type with null record;

   function Is_Update (Lock : User_Directory_Lock_Type)
                     return Boolean;

   function Is_Compatible (Lock : User_Directory_Lock_Type;
                           Other_Lock : User_Directory_Lock_Type)
                          return Boolean;

   function Is_Update (Lock : User_Directory_Lock_Type)
                      return Boolean is
   begin
      return Lock.Kind = Register;
   end Is_Update;

   function Is_Compatible (Lock : User_Directory_Lock_Type;
                           Other_Lock : User_Directory_Lock_Type)
                          return Boolean is
   begin
      case Lock.Kind is
         when Get_All_Users | Get_User =>
            return Other_Lock.Kind /= Register;
         when Register =>
            return Other_Lock.Kind = Register;
      end case;
   end Is_Compatible;

   My_Get_User_Lock : Lock_Ref := new User_Directory_Lock_Type (Get_User);
   My_Register_Lock : Lock_Ref := new User_Directory_Lock_Type (Register);
   My_Get_All_Users_Lock : Lock_Ref
     := new User_Directory_Lock_Type (Get_All_Users);

   -------------------------------------------------------------------
   --  operation for user_Directory_type Register_User

   type Register_User_Operation_Type (N, M : Natural) is
     new Normal_Operation_Type with record
        Name : String (1 .. N);
        Password : String (1 .. M);
        Params : Non_Volatile_Params_Ref;
     end record;

   procedure Do_Operation (Operation : in out Register_User_Operation_Type;
                           Data_Object : in Data_Ref);

   procedure Undo_Operation (Operation : in out Register_User_Operation_Type;
                             Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : Register_User_Operation_Type)
     return Lock_Ref;

   function Is_Update (Operation : Register_User_Operation_Type)
     return Boolean;

   --  Body part

   procedure Do_Operation (Operation : in out Register_User_Operation_Type;
                           Data_Object : in Data_Ref) is
   begin
      Register_User (To_User_Directory_Ref (Data_Object).all,
                     Operation.Name, Operation.Password,
                     Operation.Params.all);
   end Do_Operation;

   procedure Undo_Operation (Operation   : in out Register_User_Operation_Type;
                             Data_Object : in Data_Ref) is
   begin
      Remove_User (To_User_Directory_Ref (Data_Object).all, Operation.Name);
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : Register_User_Operation_Type)
     return Lock_Ref is
   begin
      return My_Register_Lock;
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : Register_User_Operation_Type)
     return Boolean is
   begin
      return True;
   end Is_Update;

   -------------------------------------------------------------------
   --  operation for user_Directory_type Get_User

   type Get_User_Operation_Type (N, M : Natural) is
     new Normal_Operation_Type with record
        Name : String (1 .. N);
        Password : String (1 .. M);
        Result : Transactional_User_Ref;
     end record;

   procedure Do_Operation (Operation : in out Get_User_Operation_Type;
                           Data_Object : in Data_Ref);

   procedure Undo_Operation (Operation : in out Get_User_Operation_Type;
                             Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : Get_User_Operation_Type)
     return Lock_Ref;

   function Is_Update (Operation : Get_User_Operation_Type)
     return Boolean;

   --  Body part

   procedure Do_Operation (Operation : in out Get_User_Operation_Type;
                           Data_Object : in Data_Ref) is
   begin
      Operation.Result := Get_User (To_User_Directory_Ref (Data_Object).all,
                                    Operation.Name, Operation.Password);
   end Do_Operation;

   procedure Undo_Operation (Operation   : in out Get_User_Operation_Type;
                             Data_Object : in Data_Ref) is
   begin
      null;
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : Get_User_Operation_Type)
     return Lock_Ref is
   begin
      return My_Get_User_Lock;
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : Get_User_Operation_Type)
     return Boolean is
   begin
      return False;
   end Is_Update;
      -------------------------------------------------------------------
   --  operation for user_Directory_type Get_All_Users
   type String_Array_Ref is access String_Array_Type;

   type Get_All_Users_Operation_Type is
     new Normal_Operation_Type with record
        Result : String_Array_Ref := null;
     end record;

   procedure Do_Operation (Operation : in out Get_All_Users_Operation_Type;
                           Data_Object : in Data_Ref);

   procedure Undo_Operation (Operation : in out Get_All_Users_Operation_Type;
                             Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : Get_All_Users_Operation_Type)
     return Lock_Ref;

   function Is_Update (Operation : Get_All_Users_Operation_Type)
     return Boolean;

   --  Body part

   procedure Do_Operation (Operation : in out Get_All_Users_Operation_Type;
                           Data_Object : in Data_Ref) is
   begin
      Operation.Result := new String_Array_Type'
        (Get_All_Users (To_User_Directory_Ref (Data_Object).all));
   end Do_Operation;

   procedure Undo_Operation (Operation   : in out Get_All_Users_Operation_Type;
                             Data_Object : in Data_Ref) is
   begin
      null;
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : Get_All_Users_Operation_Type)
     return Lock_Ref is
   begin
      return My_Get_All_Users_Lock;
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : Get_All_Users_Operation_Type)
     return Boolean is
   begin
      return False;
   end Is_Update;

   -------------------------------------------------------------------
   --  operations of the proxy

   procedure Free is new Ada.Unchecked_Deallocation
     (Non_Volatile_Params_Type'Class, Non_Volatile_Params_Ref);

   procedure Register_User
     (User_Dir : in out Transactional_User_Directory_Type;
      Name : String;
      Password : String;
      Params : Non_Volatile_Params_Type'Class) is
      My_Register_User_Operation :  Register_User_Operation_Type
        (Name'Length, Password'Length);
   begin
      My_Register_User_Operation.Name := Name;
      My_Register_User_Operation.Password := Password;
      My_Register_User_Operation.Params := new
        Non_Volatile_Params_Type'Class' (Params);
      Atomic_Do (User_Dir.My_Atomic_Call.all, My_Register_User_Operation);
      Free (My_Register_User_Operation.Params);
   end Register_User;

   function Get_User (User_Dir : Transactional_User_Directory_Type;
                      Name : String; Password : String)
                     return Transactional_User_Ref is
      My_Get_User_Operation :  Get_User_Operation_Type
        (Name'Length, Password'Length);
   begin
      My_Get_User_Operation.Name := Name;
      My_Get_User_Operation.Password := Password;
      Atomic_Do (User_Dir.My_Atomic_Call.all, My_Get_User_Operation);
      return My_Get_User_Operation.Result;
   end Get_User;

   procedure Free is new Ada.Unchecked_Deallocation
     (String_Array_Type, String_Array_Ref);

   function Get_All_Users (User_Dir : Transactional_User_Directory_Type)
                          return String_Array_Type is
      My_Get_All_Users_Operation :  Get_All_Users_Operation_Type;
   begin
      Atomic_Do (User_Dir.My_Atomic_Call.all, My_Get_All_Users_Operation);
      declare
         Result : String_Array_Type := My_Get_All_Users_Operation.Result.all;
      begin
         Free (My_Get_All_Users_Operation.Result);
         return Result;
      end;
   end Get_All_Users;

   function Create_User_Directory (Params : in Non_Volatile_Params_Type'Class)
     return Transactional_User_Directory_Ref is
      Result : Transactional_User_Directory_Ref :=
        new Transactional_User_Directory_Type;
      My_Loading_Operation : User_Directory_Loading_Type;
      My_Creation_Operation : User_Directory_Creation_Type;
      My_Saving_Operation : User_Directory_Saving_Type;
      My_Deletion_Operation : User_Directory_Deletion_Type;
   begin
      Result.My_Memory_Object := Concrete_Memory_Object_Ref
        (Create (Params,
                 My_Creation_Operation, My_Loading_Operation,
                 My_Saving_Operation, My_Deletion_Operation));
      Result.My_Atomic_Call := new Atomic_Call_Type (Result.My_Memory_Object);
      return Result;
   end Create_User_Directory;

   function Restore_User_Directory (Params : in Non_Volatile_Params_Type'Class)
                                   return Transactional_User_Directory_Ref is
      Result : Transactional_User_Directory_Ref :=
        new Transactional_User_Directory_Type;
      My_Loading_Operation : User_Directory_Loading_Type;
      My_Saving_Operation : User_Directory_Saving_Type;
      My_Deletion_Operation : User_Directory_Deletion_Type;
   begin
      Result.My_Memory_Object := Concrete_Memory_Object_Ref
        (Restore (Params,
                 My_Loading_Operation,
                 My_Saving_Operation, My_Deletion_Operation));
      Result.My_Atomic_Call := new Atomic_Call_Type (Result.My_Memory_Object);
      return Result;
   end Restore_User_Directory;

end Transactional_User_Directories;






