-------------------------------------------------------------------------------
--                            Transactional Users (body)                     --
--                                                                           --
--                  by Xavier Caron, last modified 26.09.2000                --
--                                                                           --
--                      e-mail:  xavier.caron@epfl.ch                        --
--                                                                           --
--                                                                           --
--  The Transactional Users class contains the type for the user of the      --
--  auction system.                                                          --
-------------------------------------------------------------------------------
with Operations.Creation; use Operations.Creation;
with Operations.Loading; use Operations.Loading;
with Operations.Saving; use Operations.Saving;
with Operations.Deletion; use Operations.Deletion;
with Operations.Normal; use Operations.Normal;

with Memory_Objects; use Memory_Objects;
with Memory_Objects.Concrete; use Memory_Objects.Concrete;

with Locks; use Locks;
with Locks.Read_Write; use Locks.Read_Write;

with Procedural_Transaction_Interface;
use Procedural_Transaction_Interface;

with Streams; use Streams;

with Ada.Finalization;
with Ada.Unchecked_Conversion;
with Ada.Unchecked_Deallocation;

with Users; use Users;

package body Transactional_Users is

   -------------------------------------------------------------------
   --  type conversion

   function To_Data_Ref is
      new Ada.Unchecked_Conversion (User_Ref, Data_Ref);

   pragma Warnings (OFF);
   function To_User_Ref is
      new Ada.Unchecked_Conversion (Data_Ref, User_Ref);
   pragma Warnings (ON);

   -------------------------------------------------------------------
   --  operation for user_type creation

   type User_Creation_Type (N : Natural) is new Creation_Operation_Type
     with record
        Name : String (1 .. N);
        Account_Params : Non_Volatile_Params_Ref;
     end record;

   function Create (Operation : User_Creation_Type)
     return Data_Ref;

   function Create (Operation : User_Creation_Type)
     return Data_Ref is
      Result : User_Ref := Create_User
        (Operation.Name, Operation.Account_Params.all);
   begin
      return To_Data_Ref (Result);
   end Create;

   -------------------------------------------------------------------
   --  operation for user_type loading

   type User_Loading_Type is new Loading_Operation_Type with null record;

   procedure Load (Operation : User_Loading_Type;
                   Data_Object : Data_Ref;
                   Stream : Abstract_Stream_Ref);

   procedure Create_And_Load (Operation : User_Loading_Type;
                              Data_Object : out Data_Ref;
                              Stream : Abstract_Stream_Ref);

   procedure Create_And_Load (Operation : User_Loading_Type;
                              Data_Object : out Data_Ref;
                              Stream : Abstract_Stream_Ref) is
      Result : User_Ref := new User_Type' (User_Type'Input (Stream));
   begin
      Data_Object := To_Data_Ref (Result);
   end Create_And_Load;

   procedure Free_User is new Ada.Unchecked_Deallocation
     (User_Type, User_Ref);

   procedure Load (Operation : User_Loading_Type;
                   Data_Object : Data_Ref;
                   Stream : Abstract_Stream_Ref) is
      Temp : User_Ref := To_User_Ref (Data_Object);
   begin
      --  Temp.all := User_Type'Input (Stream);
      Load_State (Temp.all, Stream);
   end Load;

   -------------------------------------------------------------------
   --  operation for user_type saving

   type User_Saving_Type is new Saving_Operation_Type with null record;

   procedure Save (Operation : User_Saving_Type;
                   Data_Object : Data_Ref;
                   Stream : Abstract_Stream_Ref);

   procedure Save (Operation : User_Saving_Type;
                   Data_Object : Data_Ref;
                   Stream : Abstract_Stream_Ref) is
   begin
      User_Type'Output
        (Stream, To_User_Ref (Data_Object).all);
   end Save;

   -------------------------------------------------------------------
   --  operation for user_type deleting

   type User_Deletion_Type is new Deletion_Operation_Type with null record;

   procedure Delete (Operation : User_Deletion_Type;
                     Data_Object : in out Data_Ref);

   procedure Delete (Operation : User_Deletion_Type;
                     Data_Object : in out Data_Ref) is
      Tmp : User_Ref := To_User_Ref (Data_Object);
   begin
      Free_User (Tmp);
      Data_Object := null;
   end Delete;

   -------------------------------------------------------------------
   --  operation for user_type get_name

   type String_Ref is access String;

   type String_Type is new Ada.Finalization.Controlled with record
      S : String_Ref;
   end record;

   procedure Finalize (S : in out String_Type);

   procedure Free is new
     Ada.Unchecked_Deallocation (String, String_Ref);

   procedure Finalize (S : in out String_Type) is
   begin
      Free (S.S);
   end Finalize;

   My_Read_Lock : Lock_Ref := Lock_Ref (Read_Lock);

   type Get_Name_Operation_Type is new Normal_Operation_Type with record
      Name : String_Type;
   end record;

   procedure Do_Operation (Operation : in out Get_Name_Operation_Type;
                           Data_Object : in Data_Ref);

   procedure Undo_Operation (Operation : in out Get_Name_Operation_Type;
                             Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : Get_Name_Operation_Type)
     return Lock_Ref;

   function Is_Update (Operation : Get_Name_Operation_Type)
     return Boolean;

   --  Body part

   procedure Do_Operation (Operation : in out Get_Name_Operation_Type;
                           Data_Object : in Data_Ref) is
   begin
      Operation.Name.S :=
        new String' (Get_Name (To_User_Ref (Data_Object).all));
   end Do_Operation;

   procedure Undo_Operation (Operation   : in out Get_Name_Operation_Type;
                             Data_Object : in Data_Ref) is
   begin
      null;
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : Get_Name_Operation_Type)
     return Lock_Ref is
   begin
      return My_Read_Lock;
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : Get_Name_Operation_Type)
     return Boolean is
   begin
      return False;
   end Is_Update;

   -------------------------------------------------------------------
   --  operation for user_type get_account

   type Get_Account_Operation_Type is new Normal_Operation_Type with record
     Result : Transactional_Account_Ref;
   end record;

   procedure Do_Operation (Operation : in out Get_Account_Operation_Type;
                           Data_Object : in Data_Ref);

   procedure Undo_Operation (Operation : in out Get_Account_Operation_Type;
                             Data_Object : in Data_Ref);

   function Get_Kind_Of_Lock (Operation : Get_Account_Operation_Type)
     return Lock_Ref;

   function Is_Update (Operation : Get_Account_Operation_Type)
     return Boolean;

   --  Body part

   procedure Do_Operation (Operation : in out Get_Account_Operation_Type;
                           Data_Object : in Data_Ref) is
   begin
      Operation.Result := Get_Account
        (To_User_Ref (Data_Object).all);
   end Do_Operation;

   procedure Undo_Operation (Operation   : in out Get_Account_Operation_Type;
                             Data_Object : in Data_Ref) is
   begin
      null;
   end Undo_Operation;

   function Get_Kind_Of_Lock (Operation : Get_Account_Operation_Type)
     return Lock_Ref is
   begin
      return My_Read_Lock;
   end Get_Kind_Of_Lock;

   function Is_Update (Operation : Get_Account_Operation_Type)
     return Boolean is
   begin
      return False;
   end Is_Update;

   -------------------------------------------------------------------
   --  operations of the proxy

   function Get_Name (User : Transactional_User_Type)  return String is
      My_Get_Name_Operation : Get_Name_Operation_Type;
   begin
      Atomic_Do (User.My_Atomic_Call.all, My_Get_Name_Operation);
      return My_Get_Name_Operation.Name.S.all;
   end Get_Name;

   function Get_Account (User : Transactional_User_Type)
                        return Transactional_Account_Ref is

      My_Get_Account_Operation :  Get_Account_Operation_Type;
   begin
      Atomic_Do (User.My_Atomic_Call.all, My_Get_Account_Operation);
      return My_Get_Account_Operation.Result;
   end Get_Account;

   procedure Free is new Ada.Unchecked_Deallocation
     (Non_Volatile_Params_Type'Class, Non_Volatile_Params_Ref);

   function Create_User (Params : in Non_Volatile_Params_Type'Class;
                         Name : String;
                         Account_Params : in Non_Volatile_Params_Type'Class)
                        return Transactional_User_Ref is
      Result : Transactional_User_Ref := new Transactional_User_Type;
      My_Loading_Operation : User_Loading_Type;
      My_Creation_Operation : User_Creation_Type (Name'Length);
      My_Saving_Operation : User_Saving_Type;
      My_Deletion_Operation : User_Deletion_Type;
   begin
      My_Creation_Operation.Name := Name;
      My_Creation_Operation.Account_Params
        := new Non_Volatile_Params_Type'Class' (Account_Params);
      Result.My_Memory_Object := Concrete_Memory_Object_Ref
        (Create (Params,
                 My_Creation_Operation, My_Loading_Operation,
                 My_Saving_Operation, My_Deletion_Operation));
      Result.My_Atomic_Call := new Atomic_Call_Type (Result.My_Memory_Object);
      Free (My_Creation_Operation.Account_Params);
      return Result;
   end Create_User;

   function Restore_User (Params : in Non_Volatile_Params_Type'Class)
     return Transactional_User_Ref is
      Result : Transactional_User_Ref := new Transactional_User_Type;
      My_Loading_Operation : User_Loading_Type;
      My_Saving_Operation : User_Saving_Type;
      My_Deletion_Operation : User_Deletion_Type;
   begin
      Result.My_Memory_Object := Concrete_Memory_Object_Ref
        (Restore (Params,
                 My_Loading_Operation,
                 My_Saving_Operation, My_Deletion_Operation));
      Result.My_Atomic_Call := new Atomic_Call_Type (Result.My_Memory_Object);
      return Result;
   end Restore_User;

end Transactional_Users;






