-------------------------------------------------------------------------------
--                             User Directories (body)                       --
--                                                                           --
--                  by Xavier Caron, last modified 29.09.2000                --
--                                                                           --
--                      e-mail:  xavier.caron@epfl.ch                        --
--                                                                           --
--                                                                           --
--  The Users Directory class contains the type for a list of all users      --
--  of the auction system.                                                   --
-------------------------------------------------------------------------------
with Ada.Unchecked_Deallocation;

package body User_Directories is

   procedure Free is new
     Ada.Unchecked_Deallocation (String, String_Ref);

   procedure Finalize (S : in out String_Type) is
   begin
      Free (S.S);
   end Finalize;

   procedure Free is new Ada.Unchecked_Deallocation (User_Node_Type,
                                                     User_Node_Ref);

   procedure Finalize (User_Dir : in out User_Directory_Type) is
      Tmp : User_Node_Ref := User_Dir.My_List;
      Tmp2 : User_Node_Ref := User_Dir.My_List;
   begin
      while Tmp /= null loop
         Tmp2 := Tmp;
         Tmp := Tmp.Next;
         Free (Tmp2);
      end loop;
   end Finalize;

   procedure Register_User (User_Dir : in out User_Directory_Type;
                            Name : String;
                            Password : String;
                            Params : Non_Volatile_Params_Type'Class) is
      Tmp : User_Node_Ref := User_Dir.My_List;
      New_User : User_Node_Ref;
      Registered : Boolean := False;
   begin
      while Tmp /= null loop
         if Tmp.My_Name.S.all = Name then
            raise User_Already_Existant;
         end if;
         Tmp := Tmp.Next;
      end loop;

      New_User := new User_Node_Type;
      New_User.My_Name.S := new String' (Name);
      New_User.My_Password.S := new String' (Password);
      New_User.My_Params := new Non_Volatile_Params_Type'Class' (Params);
      Tmp := User_Dir.My_List;

      if Tmp = null then  --  If list empty
         User_Dir.My_List := New_User;
      else
         if Tmp.My_Name.S.all > Name then  --  If at first place
            New_User.Next := Tmp;
            User_Dir.My_List := New_User;
         else
            while Tmp.Next /= null loop   --  Looking for right place.
               if Tmp.Next.My_Name.S.all > Name then
                  New_User.Next := Tmp.Next;
                  Tmp.Next := New_User;
                  Registered := True;
                  exit;
               end if;
               Tmp := Tmp.Next;
            end loop;
            if not Registered then   --  If at last place
               Tmp.Next := New_User;
            end if;
         end if;
      end if;
      User_Dir.Count := User_Dir.Count + 1;
   end Register_User;

   procedure Remove_User (User_Dir : in out User_Directory_Type;
                          Name : String) is
      Tmp : User_Node_Ref := User_Dir.My_List;
      Tmp2 : User_Node_Ref;
      Found : Boolean := False;
   begin
      if Tmp = null then
         raise User_Not_Existant;
      end if;
      if Tmp.My_Name.S.all = Name then
         User_Dir.My_List := Tmp.Next;
         Free (Tmp);
         Found := True;
      else
         Tmp2 := Tmp.Next;
         while Tmp2 /= null loop
            if Tmp2.My_Name.S.all = Name then
               Tmp.Next := Tmp2.Next;
               Free (Tmp2);
               Found := True;
               exit;
            end if;
            Tmp := Tmp2;
            Tmp2 := Tmp2.Next;
         end loop;
      end if;
      if not Found then
         raise User_Not_Existant;
      else
         User_Dir.Count := User_Dir.Count - 1;
      end if;
   end Remove_User;

   function Get_User (User_Dir : User_Directory_Type;
                      Name : String; Password : String)
                     return Transactional_User_Ref is
      Tmp : User_Node_Ref := User_Dir.My_List;
   begin
      while Tmp /= null loop
         if Tmp.My_Name.S.all = Name then
            if  Tmp.My_Password.S.all = Password then
               return Restore_User (Tmp.My_Params.all);
            else
               raise Login_Incorrect;
            end if;
         end if;
         Tmp := Tmp.Next;
      end loop;
      raise Login_Incorrect;
   end Get_User;

   function Get_All_Users (User_Dir : User_Directory_Type)
                          return String_Array_Type is
      Tmp : User_Node_Ref := User_Dir.My_List;
      Result : String_Array_Type (1 .. User_Dir.Count);
   begin
      for X in 1 .. User_Dir.Count loop
         Result (X) := Tmp.My_Name.S;
         Tmp := Tmp.Next;
      end loop;
      return Result;
   end Get_All_Users;

   procedure Load_Directory (User_Dir : out User_Directory_Type;
                             Stream : Abstract_Stream_Ref) is
      Count : Natural;
      Tmp : User_Node_Ref := User_Dir.My_List;
      First : Boolean := True;
      My_User : User_Node_Ref;
   begin
      Natural'Read (Stream, Count);
      for X in 1 .. Count loop
         My_User := new User_Node_Type;
         My_User.My_Name.S := new String' (String'Input (Stream));
         My_User.My_Password.S := new String' (String'Input (Stream));
         My_User.My_Params := new Non_Volatile_Params_Type'Class'
           (Non_Volatile_Params_Type'Class'Input (Stream));
         if First then
            User_Dir.My_List := My_User;
            First := False;
         else
            Tmp.Next := My_User;
         end if;
         Tmp := My_User;
      end loop;
      User_Dir.Count := Count;
   end Load_Directory;

   procedure Save_Directory (User_Dir : in User_Directory_Type;
                             Stream : Abstract_Stream_Ref) is
      Tmp : User_Node_Ref := User_Dir.My_List;
   begin
      Natural'Write (Stream, User_Dir.Count);
      for Count in 1 .. User_Dir.Count loop
         String'Output (Stream, Tmp.My_Name.S.all);
         String'Output (Stream, Tmp.My_Password.S.all);
         Non_Volatile_Params_Type'Class'Output (Stream, Tmp.My_Params.all);
         Tmp := Tmp.Next;
      end loop;
   end Save_Directory;

end User_Directories;

